/*
 * siman_mp.c
 *
 * This program does simulated annealing on a cluster using GSL's
 * serial implementation and MPI.  

 * The method is very simple: It does a large number of independent
 * optimizations using simulated annealing with the same starting
 * point but different random seeds.  As long as the initial
 * temperature and step size are high, this is very similar to
 * starting each optimization at a random point in the space.
 *
 * The objective function and other annealing parameters are specified
 * exactly as they are when using the GSL serial implementation.
 *
 * When estsa_mp starts, the 0th process is the master and the rest
 * are used as slaves.  The master simply sends the seeds out to the
 * slaves and gets back results.  The slaves write a log file called
 * slave-SEED.log for each seed they are asked to process.
 * 
 * When the individual optimizations are complete, the main program
 * sorts the results and reports the best one.
 * 
 * To run with mpich on each cpu listed in ~/machines:
 *
 *       % mpirun -machinefile ~/machines -np NUMPROCESSORS siman_mp NSEEDS
 *
 * NSEEDS can be larger than NUMPROCESSORS.
 * 
 * The objective function is the same as that used in siman_test.c.
 * The argument chosen to minimize the energy function takes a vector
 * argument instead of a scalar.  Each dimension is a damped sine
 * wave, and the total energy is the sum of the energy in each
 * dimension.  Like the scalar version, there are many local minima,
 * but only one global minimum.  The minimum vector is ~1.3627 in each
 * dimension and the starting point is 15.5 in each dimension.
 *
 * NOTE: Doing 10 iterations for each T enough when minimizing in
 * 2-dimensional space, but you have to use a lot of seeds.  On my
 * cluster, it got very close to global minimum on the 64th seed.
 *
 * NOTE 2: The code is a little verbose, but it gives you something to
 * look at while it's working.  And it can also help you figure out
 * what's going on if you've got bugs in your code or you have a flaky
 * network.
 *
 * NOTE 3: I'm not sure why, but the wallclock times are not reported
 * correctly.
 *
 * DMM 9/1/2007 (dmckee@ucla.edu)
 */

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <float.h>
#include <gsl/gsl_siman.h>
#include <mpi.h>

#define MYTIME (((double)clock())/CLOCKS_PER_SEC);
#define SQUARE(x) ((x)*(x))
#define MAX_FILENAME 256

/* How many dimensions in a point? That is, how many "parameters" are
 * we choosing to minimize the objective function?
*/

#define N_DIMENSIONS 2

static int G_fevals = 0; /* Tracks number of function evaluations */

/* MPI defs: */

#define WORKTAG 1
#define DIETAG 2

/* Messages to the slaves are just single integers encoding the seed 
 *
 * Messages passed back as arrays of doubles which I call "results".
 * They have the following structure:
 */

#define RES_seed(r) ((r)[0])
#define RES_fevals(r) ((r)[1])
#define RES_etime(r) ((r)[2])
#define RES_x_start(r) ((r)+3)
#define RES_val_start(r) ((r)[3+N_DIMENSIONS])
#define RES_x_end(r) ((r)+4+N_DIMENSIONS)
#define RES_val_end(r) ((r)[4+2*N_DIMENSIONS])
#define RES_LENGTH (5 + 2*N_DIMENSIONS)

typedef struct {  /* Kind of ugly, but this lets me keep track of
		     which answers came from annealing with which
		     random seeds. */
    int seed;
    double val_end;
} answer;

     
/* set up parameters for this simulated annealing run: */
     
/* how many points do we try before stepping */
#define N_TRIES 200             
     
/* how many iterations for each T? */
#define ITERS_FIXED_T 10
     
/* max step size in random walk */
#define STEP_SIZE 1.0            
     
/* Boltzmann constant */
#define K 1.0                   
     
/* initial temperature */
#define T_INITIAL 0.008         
     
/* damping factor for temperature */
#define MU_T 1.003              
#define T_MIN 2.0e-6

static gsl_siman_params_t params 
= {N_TRIES, ITERS_FIXED_T, STEP_SIZE,
   K, T_INITIAL, MU_T, T_MIN};
     
/* The objective function: What are we minimizing? */
static double E1(void *xp)
{
    double *x = (double *) xp;
    double energy = 0.0;
    int i;
    for (i=0; i<N_DIMENSIONS; i++) {
        energy += exp(-SQUARE(x[i]-1.0))*sin(8*x[i]);
    }
    G_fevals++;
    return energy;
}
     
/* The distance metric:  How far away from each other are two points? */
static double M1(void *xp, void *yp)
{
    double *x = (double *) xp;
    double *y = (double *) yp;

    double sum = 0.0;
    int i;
    for (i=0; i<N_DIMENSIONS; i++) {
        sum += SQUARE(x[i] - y[i]);
    }
     
    return sqrt(sum);
}

/* The step function: Take a step in a random direction.  The expected
   magnitude should be increasing in step_size. */

static void S1(const gsl_rng * r, void *xp, double step_size)
{
    double *old_x = (double *) xp;
    double new_x[N_DIMENSIONS];
    int i;

    for (i=0; i<N_DIMENSIONS; i++) {
        double u = gsl_rng_uniform(r);
        new_x[i] = u * 2 * step_size - step_size + old_x[i];
    }
     
    memcpy(xp, &new_x, sizeof(new_x));
}

/* The print function: Print a point. */
static void P1(void *xp)
{
    double *x = (double *) xp;
    int i;
    printf(" [");
    for (i=0; i<N_DIMENSIONS; i++) {
        printf ("%9.5g ", x[i]);
    }
    printf("]");
}
     
static int
lo2hi(const void *vp0, const void *vp1)
{
    const answer *ap0 = (const answer *)vp0;
    const answer *ap1 = (const answer *)vp1;
    double d0 = ap0->val_end;
    double d1 = ap1->val_end;
    if (d0<d1) {
        return -1;
    } else if (d0>d1) {
        return 1;
    } else {
        return 0;
    }
}

static void
master (int nseeds, answer *allanswers,
	double *bestresult, double *ptotalfevals, double *ptotaltime)
{
    int nprocesses, i;
    double *result;
    MPI_Status status;

    printf("master(0): starting\n");

    result = (double *) malloc(RES_LENGTH*sizeof(double));

    RES_val_end(bestresult) = DBL_MAX;
    *ptotalfevals=0;
    *ptotaltime=0;
  
    /* Find out how many processes there are in the default
       communicator */
  
    MPI_Comm_size(MPI_COMM_WORLD, &nprocesses);
  
    /* Seed the slaves; send one unit of work to each slave,
       assuming of course that we have enough work! */
  
    for (i = 1; (i < nprocesses) && (i <=  nseeds);  i++) {
        printf("master(0): sending message %d\n",i);
        MPI_Send(&i,             /* message buffer */
                 1,              /* one data item */
                 MPI_INT,        /* data item is an integer */
                 i,              /* destination process rank */
                 WORKTAG,        /* user chosen message tag */
                 MPI_COMM_WORLD);/* default communicator */
    }

    /* If there's work left, send it out as we get answers back: */

    if (nseeds > (nprocesses-1)) {
        for (;i <= nseeds; i++) {

            /* Receive result from a slave */

            printf("master(0):receiving message...\n");
            MPI_Recv(result,            /* message buffer */
                     RES_LENGTH,        /* one data item */
                     MPI_DOUBLE,        /* of type double real */
                     MPI_ANY_SOURCE,    /* receive from any sender */
                     MPI_ANY_TAG,       /* any type of message */
                     MPI_COMM_WORLD,    /* default communicator */
                     &status);          /* info about the received message */
            printf("master(0): message received.\n");
            (allanswers+((int) RES_seed(result))-1)->seed=(int) RES_seed(result);
            (allanswers+((int) RES_seed(result))-1)->val_end=RES_val_end(result);

            if (RES_val_end(result)<RES_val_end(bestresult)) {
                memcpy(bestresult,result,RES_LENGTH*sizeof(double));
            }
            (*ptotalfevals) +=  RES_fevals(result);
            (*ptotaltime) += RES_etime(result);
      
            /* Send the slave a new work unit */

            printf("master(0): sending message %d\n",i);
            MPI_Send(&i,             /* message buffer */
                     1,                 /* one data item */
                     MPI_INT,           /* data item is an integer */
                     status.MPI_SOURCE, /* to who we just received from */
                     WORKTAG,           /* user chosen message tag */
                     MPI_COMM_WORLD);   /* default communicator */
        }
    }

    /* There's no more work to be done, so receive all the outstanding
       results from the slaves. */
  
    for (i = 1; (i < nprocesses) && (i <= nseeds); i++) {
        printf("master(0): receiving outstanding message...\n");
        MPI_Recv(result, RES_LENGTH, MPI_DOUBLE, MPI_ANY_SOURCE,
                 MPI_ANY_TAG, MPI_COMM_WORLD, &status);
        printf("master(0): message received.\n");

        (allanswers+((int) RES_seed(result))-1)->seed=(int) RES_seed(result);
        (allanswers+((int) RES_seed(result))-1)->val_end=RES_val_end(result);
    
        if (RES_val_end(result)<RES_val_end(bestresult)) {
            memcpy(bestresult,result,RES_LENGTH*sizeof(double));
        }
        (*ptotalfevals) += RES_fevals(result);
        (*ptotaltime) += RES_etime(result);
    }

    /* Tell all the slaves to exit by sending an empty message with the
       DIETAG. */
  
    for (i = 1; i < nprocesses; i++) {
        printf("master(0): Killing %d\n",i);
        MPI_Send(0, 0, MPI_INT, i, DIETAG, MPI_COMM_WORLD);
    }

    qsort(allanswers,nseeds,sizeof(answer),lo2hi);

    free(result);
}


static void 
slave (int myrank, double *x_initial)
{
    int seed;
    double result[RES_LENGTH];
    double t0, t1;
    MPI_Status status;
    char logfilename[MAX_FILENAME];
    FILE *logfp, *stdout_orig;

    const gsl_rng_type *T;
    gsl_rng *r;

    printf("slave(%d): Starting.\n",myrank);
  
    gsl_rng_env_setup();  /* Only pay attention to the type; seed manually */

    T = gsl_rng_default;
    r = gsl_rng_alloc(T);

    while (1) {

        /* Receive a message from the master */

        printf("slave(%d) listening...\n",myrank);

        MPI_Recv(&seed, 1, MPI_INT, 0, MPI_ANY_TAG,
                 MPI_COMM_WORLD, &status);

        /* Check the tag of the received message. */

        if (status.MPI_TAG == DIETAG) {
            gsl_rng_free(r);
            return;
        }

        printf("slave(%d) received message %d\n",myrank,seed);

        /* Do the work */

	sprintf(logfilename,"slave-%d.log", seed);
	logfp = fopen(logfilename, "w");
	if (logfp == NULL) {
	  printf("Could not open %s\n", logfilename);
	  gsl_rng_free(r);
	  return;
	}

        RES_seed(result)=seed;
        gsl_rng_set(r,seed); /* seed the random number generator */
    
        memcpy(RES_x_start(result), x_initial, N_DIMENSIONS*sizeof(double));

	RES_val_start(result) = E1(x_initial);
        fprintf(logfp, "Starting from seed %d: %g",seed, RES_val_start(result));
	
        /* The annealor writes the answer into the same place it gets the starting value,
           so I initialize the result to the starting value:
        */
        memcpy(RES_x_end(result),RES_x_start(result),N_DIMENSIONS*sizeof(double));

        t0=MYTIME;
	G_fevals = 0;

	stdout_orig = stdout;
	stdout = logfp;

        gsl_siman_solve(r, RES_x_end(result), E1, S1, M1, P1,
			NULL, NULL, NULL, 
                        N_DIMENSIONS*sizeof(double), 
			params);

        stdout = stdout_orig;			
        printf("slave(%d): Done with gsl_siman_solve(%d)\n", myrank, seed);
        RES_fevals(result) = G_fevals;
        RES_val_end(result) = E1(RES_x_end(result));
        t1=MYTIME;
        RES_etime(result) = t1-t0;

	fclose(logfp);

        /* Send the result back */
        printf("slave (%d): sending result back for seed %d\n",myrank, seed);
        MPI_Send(result, RES_LENGTH, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
    }
}

int
main(int argc, char *argv[])
{
    int nseeds;
    int myrank;
    double wt0, wt1;
    int minutes;
    double walltime;
    
    wt0 = MYTIME;

    /* Initialize MPI */
    
    MPI_Init(&argc, &argv);
    
    /* Find out my identity in the default communicator */
    
    MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
    
    printf("main(%d): Starting.\n",myrank);
    
    if (argc != 2) {
        if (myrank==0) {
            printf("Usage: %s NSEEDS\n", argv[0]);
        }
        MPI_Finalize();
        return -1;
    }

    nseeds = atoi(argv[1]);

    if (myrank != 0) { /* I'm a slave! */
	double x_initial[N_DIMENSIONS];
	int i;
	for (i=0; i<N_DIMENSIONS; i++) {
	    x_initial[i] = 15.5;
	}
        printf("Calling slave(%d)...\n", myrank);

        slave(myrank, x_initial);

        /* Shut down MPI */
        MPI_Finalize();
    } else {              /* I'm the master! */
        answer *allanswers;
        double *bestresult;
        double totalfevals;
        double totaltime;
	int i;

        bestresult = (double *) malloc(RES_LENGTH*sizeof(double));
        allanswers= (answer *) malloc(nseeds*(sizeof(answer)));
        master(nseeds, allanswers, bestresult, &totalfevals, &totaltime);

        /* Shut down MPI */
        MPI_Finalize();

	/* Print results */

	printf("The answers for each seed:\n");
	for (i=0; i<nseeds; i++) {
	    printf("%d  %g\n",
		   (allanswers[i]).seed, (allanswers[i]).val_end);
	}
        free(allanswers);

	printf("\nFinal best answer:\n");
	printf("  Seed: %g\n", RES_seed(bestresult));
	printf("  Fevals: %g\n", RES_fevals(bestresult));
	printf("  X: "); P1(RES_x_end(bestresult));
	printf("  Val: %g\n", RES_val_end(bestresult));
	printf("  Total fevals: %g\n", totalfevals);
	minutes = totaltime/60;
	printf("  Total time: %dm %0.3fs\n", minutes, totaltime-60*minutes);
	wt1 = MYTIME;
        walltime = wt1 - wt0;
	minutes = walltime/60;
	printf("  Wall clock time: %dm %0.3fs\n", minutes, walltime-60*minutes);

    }

    return 0;
}

