/*
 * This code is based on the "trivial" simulated annealing example
 * in the GNU Scientific Library Reference manual:
 *
 *  http://www.gnu.org/software/gsl/manual/html_node/Trivial-example.html
 * 
 * I've extended this example so that the argument chosen to minimize
 * the energy function takes a vector argument instead of a scalar.
 * Each dimension is a damped sine wave, and the total energy is the
 * sum of the energy in each dimension.  Like the scalar version,
 * there are many local minima, but only one global minimum.  The
 * minimum vector is ~1.3627 in each dimension and the starting point
 * is 15.5 in each dimension.
 *
 * Note that doing 10 iterations for each T is not enough when
 * minimizing in 2-dimensional space.
 *
 * DMM 9/1/2007
 */

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <gsl/gsl_siman.h>

#define SQUARE(x) ((x)*(x))
     
/* set up parameters for this simulated annealing run */
     
/* how many points do we try before stepping */
#define N_TRIES 200             
     
/* how many iterations for each T? */
#define ITERS_FIXED_T 100
     
/* max step size in random walk */
#define STEP_SIZE 1.0            
     
/* Boltzmann constant */
#define K 1.0                   
     
/* initial temperature */
#define T_INITIAL 0.008         
     
/* damping factor for temperature */
#define MU_T 1.003              
#define T_MIN 2.0e-6

static gsl_siman_params_t params 
= {N_TRIES, ITERS_FIXED_T, STEP_SIZE,
   K, T_INITIAL, MU_T, T_MIN};
     
/* How many dimensions in a point? */
#define N_DIMENSIONS 2

/* The objective function: What are we minimizing? */
static double E1(void *xp)
{
    double *x = (double *) xp;
    double energy = 0.0;
    int i;
    for (i=0; i<N_DIMENSIONS; i++) {
        energy += exp(-SQUARE(x[i]-1.0))*sin(8*x[i]);
    }
    return energy;
}
     
/* The distance metric:  How far away from each other are two points? */
static double M1(void *xp, void *yp)
{
    double *x = (double *) xp;
    double *y = (double *) yp;

    double sum = 0.0;
    int i;
    for (i=0; i<N_DIMENSIONS; i++) {
        sum += SQUARE(x[i] - y[i]);
    }
     
    return sqrt(sum);
}

/* The step function: Take a step in a random direction.  The expected
   magnitude should be increasing in step_size. */

static void S1(const gsl_rng * r, void *xp, double step_size)
{
    double *old_x = (double *) xp;
    double new_x[N_DIMENSIONS];
    int i;

    for (i=0; i<N_DIMENSIONS; i++) {
        double u = gsl_rng_uniform(r);
        new_x[i] = u * 2 * step_size - step_size + old_x[i];
    }
     
    memcpy(xp, &new_x, sizeof(new_x));
}

/* The print function: Print a point. */
static void P1(void *xp)
{
    double *x = (double *) xp;
    int i;
    printf(" [");
    for (i=0; i<N_DIMENSIONS; i++) {
        printf ("%9.5g ", x[i]);
    }
    printf("]");
}
     
int
main(int argc, char *argv[])
{
    const gsl_rng_type * T;
    gsl_rng * r;
     
    double x_initial[N_DIMENSIONS];
    int i;
    for (i=0; i<N_DIMENSIONS; i++) {
        x_initial[i] = 15.5;
    }
     
    gsl_rng_env_setup();
     
    T = gsl_rng_default;
    r = gsl_rng_alloc(T);
     
    gsl_siman_solve(r, &x_initial, E1, S1, M1, P1,
                    NULL, NULL, NULL, 
                    N_DIMENSIONS*sizeof(double), params);
     
    gsl_rng_free (r);
    return 0;
}

